// Game:           Prince of Persia (SNES)
// Category:       100%
// Emulator:       snes9x-1.60-win32-x64
// Original by:    David from forum.princed.org
// Updated:        2023-01-08
// IGT timing:     YES

// Base memory path:
// snes9x-1.60-win32-x64: "snes9x-x64.exe"+008D8C38

// Open questions:
// * What if the player does the actions in different order?
// * What if the player (dies and) restarts a level?

state("snes9x-x64")
{
    byte720 LevelObjects          : "snes9x-x64.exe", 0x008D8C38, 0x1E170; // 7F:E170
    byte NextLevel                : "snes9x-x64.exe", 0x008D8C38, 0x0544; // 7E:0544
    byte CurrentLevel             : "snes9x-x64.exe", 0x008D8C38, 0x0579; // 7E:0579; 0 means LEVEL 1, and so on.
    byte CharScrn                 : "snes9x-x64.exe", 0x008D8C38, 0x0462; // 7E:0462
    byte KidMaxHP                 : "snes9x-x64.exe", 0x008D8C38, 0x050A; // 7E:050A
    ushort FrameCount             : "snes9x-x64.exe", 0x008D8C38, 0x052D; // 7E:052D
    byte FrameCountHalf           : "snes9x-x64.exe", 0x008D8C38, 0x052F; // 7E:052F
    byte StartRoom                : "snes9x-x64.exe", 0x008D8C38, 0x1F110; // 7F:F110

    sbyte Guard_alive             : "snes9x-x64.exe", 0x008D8C38, 0x0486; // 7E:0486
    byte Guard_room               : "snes9x-x64.exe", 0x008D8C38, 0x0482; // 7E:0482
    byte Guard_direction          : "snes9x-x64.exe", 0x008D8C38, 0x047A; // 7E:047A
    byte Guard_hp                 : "snes9x-x64.exe", 0x008D8C38, 0x050B; // 7E:050B
    byte Guard_frame              : "snes9x-x64.exe", 0x008D8C38, 0x0477; // 7E:0477
    ushort Guard_sequence_pointer : "snes9x-x64.exe", 0x008D8C38, 0x0480; // 7E:0480
}

startup
{
	//refreshRate = 30; // Prince of Persia runs at 12 fps, let's update 2.5x as often to be sure
/*
	settings.Add("test_group", true, "Test setting group");
		settings.Add("test", false, "Test setting", "test_group");
*/
	// This is the list of actions needed for 100% completion.
	// Keep this in sync with the splits!
	vars.checks = new int[][]
	{
		// columns: type, level, room, tile
		// types:
		// 'p' = pick up (potion, sword)
		// 'g' = guard (ignores the tile)
		// 'x' = level exit (ignores the room and the tile)

		// Level 1
		new int[]{'p', 1,  8, 27}, // Level 1: Heal potion at room 8, tile 27 (attic, under the guard)
		new int[]{'p', 1, 12,  4}, // Level 1: Heal potion at room 12, tile 4 (cellar)
		new int[]{'p', 1, 20, 25}, // Level 1: Sword at room 20, tile 25
		new int[]{'p', 1, 23,  1}, // Level 1: Heal potion at room 23, tile 1 (attic left, above the skeleton)
		new int[]{'g', 1,  8, -1}, // Level 1: Guard at room 8, tile 6 (attic)
		new int[]{'g', 1,  1, -1}, // Level 1: Guard at room 1, tile 17 (before the exit)
		new int[]{'x', 1, -1, -1}, // Level 1: Level exit

		// Level 2
		new int[]{'g', 2,  7, -1}, // Level 2: Guard at room 7
		new int[]{'p', 2,  8, 14}, // Level 2: Heal potion at room 2, tile 14 (room below the guard)
		new int[]{'g', 2, 11, -1}, // Level 2: Guard at room 11 (still underground)
		new int[]{'p', 2,  5, 26}, // Level 2: Heal potion at room 5, tile 26 (above the ceiling)
		new int[]{'g', 2,  2, -1}, // Level 2: Guard at room 2 (at the skeleton)
		new int[]{'p', 2,  1, 27}, // Level 2: Heal potion at room 1, tile 27 (between the gates)
		new int[]{'g', 2,  0, -1}, // Level 2: Guard at room 0 (bluish, at the spikes)
		new int[]{'p', 2, 13, 28}, // Level 2: Life potion at room 13, tile 28
		new int[]{'p', 2, 23, 25}, // Level 2: Heal potion at room 23, tile 25 (ceiling)
		new int[]{'g', 2, 16, -1}, // Level 2: Guard at room 16 (bluish)
		new int[]{'g', 2, 15, -1}, // Level 2: Guard at room 15 (next to the exit)
		new int[]{'g', 2, 25, -1}, // Level 2: Guard at room 25 (bluish) (under the button which opens the exit)
		new int[]{'x', 2, -1, -1}, // Level 2: Level exit

		// TODO: Levels 3-20
	};

	vars.check_index = 0;
	vars.split = false;
	vars.playing = false;
	vars.reset_guards = (Action)delegate()
	{
		int[] array = new int[24];
		for (int i = 0; i < 24; i++) array[i] = i;
		vars.guard_original_rooms = array;
	};
	vars.reset_guards();
}

reset
{
	// KidMaxHP == 0 when the game is reset.
	// CurrentLevel == 0 on level 1 *and* in the main menu!
	// CurrentLevel >= 21 are the demo and training levels.
	// CurrentLevel == 128 during cutscenes.
	// CurrentLevel == 252 after the demo level was interrupted and before the main menu fully appears.
	// CurrentLevel == 254 after GAME END was selected and before the main menu fully appears.
	// CurrentLevel == 255 when restarting the level.
	bool reset = (
		current.KidMaxHP == 0
		|| (current.CurrentLevel >= 21 && current.CurrentLevel < 128)
		|| current.CurrentLevel == 252
		|| current.CurrentLevel == 254
	);
	if (reset)
	{
		print("reset");
		print("CurrentLevel = " + current.CurrentLevel);
		vars.check_index = 0;
		vars.playing = false;
		vars.reset_guards();
	}
	return reset;
}

start
{
	// This should become true only when the player selects START GAME.
	// We check StartRoom to prevent false positives, as otherwise the condition is true after the player quits the demo/training level.
	bool start = (
		current.KidMaxHP == 3
		&& current.FrameCount == 0
		&& current.CurrentLevel == 0
		&& current.StartRoom == 22
	);
	if (start)
	{
		print("start");
		vars.check_index = 0;
		vars.playing = true;
		vars.reset_guards();
	}
	return start;
}

update
{
	if (!vars.playing) return;

	// CurrentLevel == 255 when restarting the level.
	if (current.CurrentLevel == 255) vars.reset_guards();

	// CurrentLevel == 128 during cutscenes.
	if (current.CurrentLevel >= 128) return;

	//print("It's alive!");
	//print("CurrentLevel = "+current.CurrentLevel);
	//print("NextLevel = "+current.NextLevel);
	//print("KidMaxHP = "+current.KidMaxHP);

	// If a guard moved to another room:
	if (current.Guard_direction != 0x7F
		&& old.Guard_direction != 0x7F
		&& current.Guard_room != old.Guard_room
		&& current.Guard_direction == old.Guard_direction
		&& current.Guard_hp == old.Guard_hp
		// TODO: Check more attributes of the guard to make sure they are the same guard?
		// TODO: Check if the prince is in fight mode? Or, can guards move between rooms even if he isn't?
		)
	{
		// If there is a guard in room R, then guard_original_rooms[R] tells the room where that guard started the level.
		int guard_original_room = vars.guard_original_rooms[old.Guard_room];
		print("A guard moved from room "+old.Guard_room+" to "+current.Guard_room+". (original room: "+guard_original_room+")");
		vars.guard_original_rooms[current.Guard_room] = guard_original_room;
	}

	if (vars.check_index >= vars.checks.Length) return;

	int[] check = vars.checks[vars.check_index];
	char type = (char)check[0];
	int level = check[1];
	int room = check[2];
	int tile = check[3];
	//print("check_index = "+vars.check_index);
	//print("type = "+type);
	//print("level = "+level);
	//print("room = "+room);
	//print("tile = "+tile);

	// For debugging: Quickly skip over each action if the game is past its level.
	if (current.CurrentLevel+1 > level)
	{
		vars.check_index++;
		vars.split = true;
		vars.reset_guards();
		return;
	}

	// Type IDs of objects which can be picked up:
	// (Source: https://www.princed.org/wiki/SNES_format#Levels)
	// 0x0A = sword
	// 0x12 = heal potion
	// 0x13 = life potion
	// 0x14 = hurt potion
	// 0x15 = upside down potion
	// 0x16 = slow fall potion
	// 0x1A = potion of warp
	// 0x1C = kill potion

	switch (type)
	{
		case 'p': // pick up
			if (current.CurrentLevel+1 == level)
			{
				int offset = room*30+tile;
				// TODO: Compare the old value against the specific object type?
				// PROBLEM: If we check the old value in any way, then we won't detect actions if the player does them out of order.
				// PROBLEM: If we *don't* check the old value, we can get false positives before the level is fully loaded.
				if (current.LevelObjects[offset] == 0x00 && old.LevelObjects[offset] != 0x00)
				{
					//print($"The player picked up the potion at: level {level}, room {room}, tile {tile}");
					print("The player picked up the item at: level "+level+", room "+room+", tile "+tile);
					vars.check_index++;
					vars.split = true;
				}
			}
		break;

		case 'g': // guard
			if (current.Guard_direction != 0x7F) // if there is an active guard
			{
				if ((current.Guard_alive == 0 || current.Guard_frame == 184)
					&& !(old.Guard_alive == 0 || old.Guard_frame == 184))
				{
					int guard_original_room = vars.guard_original_rooms[current.Guard_room];
					print("A guard died in room " + current.Guard_room + " of level " + (current.CurrentLevel+1)
						+ ". (original room: "+guard_original_room+")");
					if (current.CurrentLevel+1 == level && /*current.Guard_room*/ guard_original_room == room)
					{
						vars.check_index++;
						vars.split = true;
					}
				}
			}
		break;

		case 'x': // level exit
			if (current.CurrentLevel+1 == level
				&& current.NextLevel != current.CurrentLevel)
			{
				vars.check_index++;
				vars.split = true;
				vars.reset_guards();
			}
		break;
	}

}

split
{
	bool split = vars.split;
	vars.split = false;
	return split;
}

gameTime
{
	// 1 minute = 0x01A9 frames, 1 second = 7 frames.
	// So 1 minute is slightly more than 60 seconds, and they don't even match real time.
	return TimeSpan.FromMinutes( ((double)current.FrameCount + 0.5 * (double)current.FrameCountHalf) / (double)0x01A9);
}

isLoading
{
	// Force Game Time to show the actual Game Time...
	return true;
}
